#include <XSFit/Fit/StatManager.h>

#include <XSContainer.h>
#include <XSstreams.h>
#include <XSFit/Fit/Fit.h>
#include <XSFit/Fit/FitMethod.h>
#include <XSModel/Data/BackCorr/Background.h>
#include <XSModel/Data/DataUtility.h>
#include <XSModel/Data/SpectralData.h>
#include <XSModel/Data/Detector/Response.h>
#include <XSModel/GlobalContainer/DataContainer.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSModel/Model/Model.h>
#include <XSModel/Model/CompCombiner.h>
#include <XSModel/Parameter/ModParam.h>
#include <XSModel/Parameter/ResponseParam.h>
#include <XSStat/StatMethod.h>
#include <XSStat/FitStatMethod.h>
#include <XSStat/TestStatMethod.h>
#include <XSStat/Weight.h>
#include <XSUtil/Error/Error.h>
#include <XSUtil/Utils/XSutility.h>
#include <iostream>
#include <iomanip>

StatManager* StatManager::s_instance = 0;
Bayes StatManager::s_bayes = Bayes();

StatManager::StatManager()
 :m_fitStatMethods(),
  m_testStatMethod(0),
  m_totalStatistic(0.0),
  m_totalTestStatistic(0.0),
  m_bayesianPrior(0.0),
  m_usingSingleStat(0),
  m_defaultStatMethod(0),
  m_alpha(),
  m_beta(),
  m_jacobianCalc(new ParDiff(), "leven")
{
}  

StatManager::~StatManager()
{
   std::map<size_t,FitStatMethod*>::const_iterator itFm = m_fitStatMethods.begin();
   std::map<size_t,FitStatMethod*>::const_iterator itFmEnd = m_fitStatMethods.end();
   while ( itFm != itFmEnd ) {
     delete itFm->second;
     itFm++;
   }
   m_fitStatMethods.clear();

   delete m_usingSingleStat;
   delete m_testStatMethod;
   delete m_defaultStatMethod;
}

StatManager* StatManager::Instance()
{
   if (s_instance == 0)
   {
      s_instance = new StatManager();
   }
   return s_instance;
}

void StatManager::Update (Subject* changed)
{

   m_totalStatistic = 0.0;
   m_totalTestStatistic = 0.0;
   m_bayesianPrior = 0.0;
   // In addition to the direct effects of the statistic command,
   // StatMethod configurations can be affected indirectly through
   // the data, model, and response commands.

   checkForSingleStat();
   setSpectraForTestStatMethods(XSContainer::fit->spectraToFit());

   // Don't let exceptions propagate from anywhere in Notify/Update chain.
   try
   {

      if (!XSContainer::fit->spectraToFit().empty() && !XSContainer::fit->activeModels().empty())
      {
         // stat calculation may throw.
         initializeStats();
         if (XSContainer::fit->renormType()==Fit::AUTO)
         {
            renormalizeStats();
            performStats();           
         }
         else
         {
            // When not auto-renorming, there's no reason to perform
            // new Model calculate and fold operations.  These should
            // already have been done in Fit::Update and should still
            // be valid.
            performStats(false);
         }
         if (!Fit::errorCalc())
            reportStats();
      }
   }
   catch (YellowAlert&)
   {
      // Any exceptions should have sent out an error message, but
      // let's make it more obvious that something is wrong.
      m_totalStatistic = -1.0;
   }
} // end StatManager::Update


void StatManager::setSpectraForTestStatMethods(const std::map<size_t,SpectralData*>& spectraToFit)
{
  // clear the spectra set for the current test statistic
  m_testStatMethod->spectraForStat().clear();

  // loop over spectra being fit adding to the list of spectra for the test statistic

   std::map<size_t,SpectralData*>::const_iterator itSpec = spectraToFit.begin();
   std::map<size_t,SpectralData*>::const_iterator itEnd = spectraToFit.end();
   while (itSpec != itEnd) {
      const SpectralData* spec = itSpec->second;
      m_testStatMethod->spectraForStat().push_back(spec);
      ++itSpec;
   }

}

void StatManager::renormalizeStats() const
{
  // Attempt a renormalization only if using a single stat.
  if (!m_usingSingleStat) return;

  // If we are missing either models or data give up.
  if (  XSContainer::models->activeModelNames().empty() || XSContainer::datasets->numberOfSpectra() == 0) {
    throw Fit::CantInitialize(" renorm: either no model defined or no data read ");
  }

  // Since we are using a single statistic just get the statistic for the first spectrum
  // to check whether we can use the statistic for renormalization.
  
  FitStatMethod* statMeth = getStatMethodForSpectrum(XSContainer::fit->spectraToFit().begin()->first);

  if (statMeth->name() == "cstat" || 
      statMeth->name() == "lstat" || 
      statMeth->name() == "pgstat") {
    // These stat methods aren't set up for renorming with
    // background data.  
    for (XSContainer::DataArrayConstIt itDs = XSContainer::datasets->dataArray().begin(); 
	 itDs != XSContainer::datasets->dataArray().end(); ++itDs) {
      const DataSet* ds = itDs->second;
      if (ds->isMultiple()) {
	XSContainer::SpectralDataMapConstIt itSd = ds->multiSpectralData().begin();
	while (itSd != ds->multiSpectralData().end()) {
	  const SpectralData* sd = itSd->second;
	  if (sd && sd->background())  return;
	  ++itSd;
	}
      } else if (const SpectralData* sd = ds->spectralData()) {
	if (sd->background())  return;
      }
    }
  }

  // first, compute and store current model.
  std::map<Model*,bool> frozenComponents;

  size_t numberOfNorms (0);
  size_t numberOfFrozenNorms(0);
           
  // If any mixing is detected, skip renorm entirely.
  // Note:  If renorming is to be reinstated at some point for
  // mixing models, must ensure that no zero-sized SumComponent
  // arrays get passed down to mix function. (This bug arose in 
  // the past when spectra with no frozen norms were bundled for 
  // mixing with those which do have them.) 
  bool anyMixing=false;           
  XSContainer::ModelMapConstIter itMod (XSContainer::models->modelSet().begin());
  XSContainer::ModelMapConstIter itModEnd (XSContainer::models->modelSet().end());
  while ( itMod != itModEnd && !anyMixing) {
    Model* currentModel = itMod->second;
    frozenComponents[currentModel] = false;
    if ( currentModel->isActive() ) { 
      if (currentModel->mixingLocations().first.size() ||
	  currentModel->mixingLocations().second.size())
	anyMixing = true;
      else {
	std::list<ModParam*> norms (currentModel->normParams());
	numberOfNorms += norms.size();
	for (std::list<ModParam*>::const_iterator np = norms.begin();
	     np != norms.end(); ++np) {
	  if ((*np)->isFrozen() || Parameter::isLinkedToFrozen(*np)) {
	    frozenComponents[currentModel] = true; 
	    ++numberOfFrozenNorms;
	  }            
	}
      }  
    }
    ++itMod;
  }

  if (anyMixing) return;
              
  if (numberOfFrozenNorms == numberOfNorms ) {
    if (!XSContainer::fit->errorCalc()) {        
      tcout << " Warning: renorm - no variable model to allow  renormalization"
	    << std::endl; 
    }   
  } else {
    std::pair<Real,Real> fSums;
    Real ofSum(0.);    // =>     fSums[0]
    Real mfSum(0.);    // =>     fSums[1]

    // reset stat method for all spectra
    std::map<size_t, SpectralData*>::const_iterator itSpec = XSContainer::fit->spectraToFit().begin();
    std::map<size_t, SpectralData*>::const_iterator itSpecEnd = XSContainer::fit->spectraToFit().end();
    while ( itSpec != itSpecEnd ) {
      getStatMethodForSpectrum(itSpec->first)->reset(itSpec->second);
      itSpec++;
    }

    std::map<Model*,ArrayContainer > saveFrozenModVals;
    std::set<string>::const_iterator itModNames = XSContainer::fit->activeModels().begin();
    std::set<string>::const_iterator itModNamesEnd = XSContainer::fit->activeModels().end();
    while (itModNames != itModNamesEnd) {                 
      std::vector<Model*> modGroups = XSContainer::models->lookupModelGroup(*itModNames);
                 
      // Only models with frozen norms are to be recalculated.
      bool anyFrozenNormsInGroup=false;
                 
      for (size_t iMod=0; iMod<modGroups.size(); ++iMod) {
	Model* mod = modGroups[iMod];
	if (frozenComponents[mod]) anyFrozenNormsInGroup = true;
      }
      for (size_t iMod=0; iMod<modGroups.size(); ++iMod) {
	Model* mod = modGroups[iMod];
	mod->fold();
	if (frozenComponents[mod]) {
	  ArrayContainer& savModVal = saveFrozenModVals[mod];
	  ArrayContainer::const_iterator it = mod->foldedModel().begin();
	  ArrayContainer::const_iterator itEnd = mod->foldedModel().end();
	  while (it != itEnd) {
	    savModVal.insert(ArrayContainer::value_type(it->first,it->second));
	    ++it;
	  }
	  mod->calcComponents(false, true);
	  std::vector<Model*> modToCombine(1, mod);
	  CompCombiner combiner(modToCombine);
	  CompCombiner::iterator itComb = combiner.begin();
	  CompCombiner::iterator itCombEnd = combiner.end();
	  while (itComb != itCombEnd) {
	    ++itComb;
	  }
	  mod->fold(); 
	}
      }
      ++itModNames;
    } // end modNames loop



    // Want to analyze spectra grouped by data group number.
    // DataContainer currently provides no ready-made way to
    // do this, so in order to avoid O(N^2) behavior let's
    // collate things here.
    const size_t nDgs = XSContainer::datasets->numberOfGroups();
    const size_t nSpecs = XSContainer::datasets->numberOfSpectra();
    std::vector<std::vector<const SpectralData*> > specsByGroup(nDgs);
    for (size_t iSp=1; iSp<=nSpecs; ++iSp) {
      const SpectralData* spec = XSContainer::datasets->lookup(iSp);
      const size_t dgNum = spec->parent()->dataGroup();
      specsByGroup[dgNum-1].push_back(spec);
    }

    for (size_t iDg=0; iDg<nDgs; ++iDg) {
      std::vector<Model*> modsForGroup(XSContainer::datasets->numSourcesForSpectra(),0);
      const std::set<size_t>& sourcesForGroup =  
	XSContainer::datasets->dgToSources().find(iDg+1)->second;
      std::set<size_t>::const_iterator itSource = sourcesForGroup.begin();
      std::set<size_t>::const_iterator itSourceEnd = sourcesForGroup.end();
      while (itSource != itSourceEnd) {
	const string modName = XSContainer::models->lookupModelForSource(*itSource);
	if (modName.length()) {
	  // Model must be active/on if in here.
	  Model* mod = XSContainer::models->lookup(modName, iDg+1);
	  modsForGroup[*itSource-1] = mod;
	}
	++itSource;
      }

      const std::vector<const SpectralData*>& specsForGroup = specsByGroup[iDg];
      for (size_t iSp=0; iSp<specsForGroup.size(); ++iSp) {
	const SpectralData* spec = specsForGroup[iSp];
	const std::vector<Response*>& detectors = spec->detector();
	std::vector<Model*> modsForSpec;
	for (size_t iDet=0; iDet<detectors.size(); ++iDet) {
	  if (detectors[iDet]) {
	    if (modsForGroup[iDet]) modsForSpec.push_back(modsForGroup[iDet]);
	  }
	}
	// adjustForFrozen functions ASSUME at least one 
	// model is applied to spec, so filter out no model 
	// case here.
	if (modsForSpec.size()) {
	  FitStatMethod* fitStat = getStatMethodForSpectrum(spec->spectrumNumber());
	  fSums = fitStat->adjustForFrozen(spec, modsForSpec, saveFrozenModVals, XSContainer::fit->spectraToFit().size());
	  ofSum += fSums.first;
	  mfSum += fSums.second;
	}
      } 
    } // end data group loop


    Real renorm(1.);
    if(std::abs(ofSum - mfSum) < SMALL) {
      if (!XSContainer::fit->errorCalc()) {
	tcout << " renorm: no renormalization necessary" <<std::endl;   
      }
    } else if(std::abs(mfSum) > SMALL) { 
      renorm = ofSum / mfSum;
      itMod = XSContainer::models->modelSet().begin();
      while (itMod != itModEnd) {
	Model* currentModel (itMod->second);
	if ( currentModel->isActive()) {
	  std::list<ModParam*> norms (currentModel->normParams());
	  std::list<ModParam*>::iterator np = norms.begin();
	  std::list<ModParam*>::iterator npEnd = norms.end();
	  while (np != npEnd) {
	    if (!(*np)->isFrozen() && !(*np)->isLinked()) {
	      Real newValue = (*np)->value()*renorm;
	      newValue = std::max( (*np)->value('l'),
				   std::min(newValue,(*np)->value('h')));
	      (*np)->setValue(newValue,'v');
	    }
	    ++np;       
	  }
	}
	++itMod;
      }
    }
  } // end frozenNorms != numNorms

}

void StatManager::initializeStats()
{
  // delete any already existing objects in the m_fitStatMethods map and clear it
  std::map<size_t, FitStatMethod*>:: const_iterator itStat = m_fitStatMethods.begin();
  std::map<size_t, FitStatMethod*>:: const_iterator itStatEnd = m_fitStatMethods.end();
  while ( itStat != itStatEnd ) {
    delete itStat->second;
    itStat++;
  }
  m_fitStatMethods.clear();

  // read the statistic information for each dataset, construct the m_fitStatMethods
  // map and initialize the objects. 
  std::map<size_t, SpectralData*>::const_iterator itSpec = XSContainer::fit->spectraToFit().begin();
  std::map<size_t, SpectralData*>::const_iterator itSpecEnd = XSContainer::fit->spectraToFit().end();
  while ( itSpec != itSpecEnd ) {
    const SpectralData* spec = itSpec->second;
    const size_t specNum = itSpec->first;
    string name, modstring;
    int numeral;
    StatMethod::interpretStatCommandName(spec->statCommandName(), name, numeral, modstring);
    setStatMethodForSpectrum(specNum, name, numeral, modstring);
    getStatMethodForSpectrum(specNum)->initialize(spec);
    itSpec++;
  }

  m_testStatMethod->initialize();
}

void StatManager::reportStats() const
{
  using namespace std;

  int totalDegrees = totalDegreesOfFreedom().first;
  if (totalDegrees < 0) {
    tcout << "Ill-formed Fit problem - number of variable parameters exceeds number of bins"
	  << endl;
    return;
  }

  if (XSContainer::fit->activeModels().empty()) {
    tcout << "No active models" << endl;
    return;
  }

  // the critical delta statistic is used in deciding how many significant figures
  // to use on the statistic value
  Real deltaCrit = abs(XSContainer::fit->fitMethod()->deltaCrit());

  ios_base::fmtflags saveFlags(tcout.flags());
  streamsize savePrec(tcout.precision());

  std::map<size_t, SpectralData*>::const_iterator itSpec = XSContainer::fit->spectraToFit().begin();
  std::map<size_t, SpectralData*>::const_iterator itSpecEnd = XSContainer::fit->spectraToFit().end();
  const size_t nSpectra = XSContainer::fit->spectraToFit().size();
  const size_t nGroups = XSContainer::datasets->numberOfGroups();

  if ( nSpectra == 1 ) {

    FitStatMethod* fitStat = getStatMethodForSpectrum(itSpec->first);
    const size_t bins = fitStat->nBinsUsed(itSpec->second);
    Real truncatedStat = StatMethod::setOutputFormat(m_totalStatistic, deltaCrit);
    tcout << "\nFit statistic  : " << setw(25) << left << fitStat->fullName()
	  << setw(10) << right << truncatedStat << "     using " << bins << " bins." << std::endl;
    fitStat->reportWarnings(itSpec->second);

  } else {

    bool first(true);
    while ( itSpec != itSpecEnd ) {
      FitStatMethod* fitStat = getStatMethodForSpectrum(itSpec->first);
      const size_t bins = fitStat->nBinsUsed(itSpec->second);
      Real truncatedStat = StatMethod::setOutputFormat(fitStat->value(), deltaCrit);
      if ( first ) {
	tcout << "\nFit statistic  : ";
      } else {
	tcout << "                 ";
      }
      tcout << setw(25) << left << fitStat->fullName() << setw(10) << right
	    << truncatedStat << "     using " << bins << " bins"
	    << ", spectrum " << itSpec->second->spectrumNumber();
      if ( nGroups > 1 ) tcout << ", group " << itSpec->second->parent()->dataGroup();
      tcout << "." << endl;
      fitStat->reportWarnings(itSpec->second);
      first = false;
      itSpec++;
    }

    Real truncatedTotal = StatMethod::setOutputFormat(m_totalStatistic, deltaCrit);
    tcout << setw(42) << left << "Total fit statistic " << setw(10) << right 
	  << truncatedTotal << "     with " << totalDegrees << " d.o.f." 
	  << std::endl;
  }

  if (m_bayesianPrior != 0.0) {
    tcout <<"   (Includes a Bayesian contribution of " <<m_bayesianPrior<<")\n";
  }
  tcout << std::endl;

  tcout << "Test statistic : ";
  const size_t bins = m_testStatMethod->nBinsUsed();
  Real truncatedStat = StatMethod::setOutputFormat(m_testStatMethod->value(), deltaCrit);
  tcout << setw(25) << left << m_testStatMethod->fullName() << setw(10) << right
	    << truncatedStat << "     using " << bins << " bins." << endl;
  m_testStatMethod->reportWarnings();

  if ( m_testStatMethod->canCalcNullHyp() ) {
    Real nullHypProb = m_testStatMethod->calcNullHyp(totalDegrees);
    if ( nullHypProb >= 0.0 ) {
      tcout << " Null hypothesis probability of " << scientific 
	    << nullHypProb << " with " << totalDegrees 
	    << " degrees of freedom" << endl;
    }
  }

  tcout.precision(savePrec);
  tcout.flags(saveFlags);
   
  if (!XSContainer::fit->isStillValid())
    tcout << " Current data and model not fit yet." << endl;         

}

void StatManager::performStats(bool doCalcFold)
{

   m_totalStatistic = 0.0;
   m_totalTestStatistic = 0.0;
   m_bayesianPrior = 0.0;
   
   if (doCalcFold)
   {
      std::set<string>::iterator ss (XSContainer::fit->activeModels().begin());
      std::set<string>::iterator ssEnd (XSContainer::fit->activeModels().end());
      while ( ss != ssEnd)
      {
 	 XSContainer::models->calculate(*ss);
	 XSContainer::models->fold(*ss);
         ++ss;
      }
   }  

   // lPrior returns 0.0 if Bayes flag is not on.
   // For exception safety concerns (in case StatMethod perform should throw),
   // let's not modify m_bayesianPrior till after we're out of stat loop.
   const Real bayesianPrior = -2.0* s_bayes.lPrior(XSContainer::fit);

   std::map<size_t, SpectralData*>::const_iterator itSpec = XSContainer::fit->spectraToFit().begin();
   std::map<size_t, SpectralData*>::const_iterator itSpecEnd = XSContainer::fit->spectraToFit().end();
   m_totalStatistic = 0.0;
   while ( itSpec != itSpecEnd ) {
     FitStatMethod* statMeth = getStatMethodForSpectrum(itSpec->first);
     statMeth->perform(itSpec->second);
     m_totalStatistic += statMeth->value();
     itSpec++;
   }

   m_bayesianPrior = bayesianPrior;
   m_totalStatistic += m_bayesianPrior;  

   m_testStatMethod->perform();
   m_totalTestStatistic = m_testStatMethod->value();

}

void StatManager::peakResidual(std::pair<Real,Real>&  max, std::pair<Real,Real>&  min, size_t spectrum, const std::pair<Real,Real>& range) const
{

  // Trap case of not fit 
  if ( XSContainer::fit->activeModels().empty() || XSContainer::fit->spectraToFit().empty()) {
    throw StatMethod::FitNotDefined(" : Peak Residual");
  }

  // find correct SpectralData*
  std::map<size_t,SpectralData*>::const_iterator sm = XSContainer::fit->spectraToFit().find(spectrum);
  if (sm == XSContainer::fit->spectraToFit().end()) 
  {
     throw YellowAlert("No response*model associated with spectrum\n");   
  }
  SpectralData* sp = sm->second;
  size_t specNum = sp->spectrumNumber();

  // get the differences between the spectrum and the model (this is statistic
  // dependent because in some cases the statistic calculation will have already
  // calculated and saved this information - in other cases it hasn't)
  RealArray residual(getStatMethodForSpectrum(specNum)->getDifferences(sp));

  const std::valarray<size_t>& IN = sp->indirectNotice();
  int M (IN.size());
  int lower(0);
  int upper(M-1);
  if ( range.first != 0 && range.second != 0)
  {
     sp->energyChannel(range.first,lower);
     sp->energyChannel(range.second,upper);
  }
  Real maxResid (-LARGE);
  Real maxEnergy (0);


  Real minResid (LARGE);
  Real minEnergy (0);
  for (int j = lower; j <= upper; ++j)
  {
     Real resid = residual[j];
     if (resid > maxResid) 
     {
        maxResid = resid;
        maxEnergy = sp->channelEnergy(j);
     }         
     if (resid < minResid) 
     {
        minResid = resid;
        minEnergy = sp->channelEnergy(j);
     } 
  }

  int nResps = 0;
  Real minEffArea = 0.0;
  Real maxEffArea = 0.0;     
  for (size_t i=0; i<sp->detector().size(); ++i)
  {
     const Response* rsp = sp->detector(i);
     if (rsp)
     {
        int idx;
        XSutility::find(rsp->energies(), maxEnergy, idx);
        const RealArray& effectiveArea = rsp->efficiency();
        if (idx < 0 || static_cast<size_t>(idx) >= effectiveArea.size())
        {
           // This should never happen
           throw YellowAlert("Peak resid energy is outside effective area array\n");
        }
        maxEffArea += effectiveArea[idx];
        XSutility::find(rsp->energies(), minEnergy, idx);
        if (idx < 0 || static_cast<size_t>(idx) >= effectiveArea.size())
        {
           // This should never happen
           throw YellowAlert("Peak resid energy is outside effective area array\n");
        }
        minEffArea += effectiveArea[idx];
        ++nResps;
     }
  }
  // nResps must be > 0 or it would have been caught above.
  maxEffArea /= static_cast<Real>(nResps);
  if (maxEffArea == 0.0)
     throw YellowAlert("Zero effective area for bin: Cannot calculate peak strength.\n");
  minEffArea /= static_cast<Real>(nResps);
  if (minEffArea == 0.0)
     throw YellowAlert("Zero effective area for bin: Cannot calculate peak strength.\n");
  max.first = maxResid/maxEffArea;
  max.second = maxEnergy;
  min.first = minResid/minEffArea;
  min.second = minEnergy;

}

void StatManager::checkForSingleStat()
{
   std::map<size_t, SpectralData*>::const_iterator itSpec = XSContainer::fit->spectraToFit().begin();
   std::map<size_t, SpectralData*>::const_iterator itSpecEnd = XSContainer::fit->spectraToFit().end();

   if (m_usingSingleStat)
      delete m_usingSingleStat;
   m_usingSingleStat = 0;
   bool isFirst=true;
   bool foundMulti=false;
   string singleCommandName;
   while (!foundMulti && itSpec != itSpecEnd)
   {
      const SpectralData* spec = itSpec->second;
      if (isFirst)
      {
         singleCommandName = spec->statCommandName();
         isFirst = false;
      }
      else
      {
         if (spec->statCommandName() != singleCommandName)
         {
            foundMulti = true;
         }
      }
      ++itSpec;
   }
   if (!foundMulti && singleCommandName.length())
   {
      string name, modifier;
      int number=0;
      StatMethod::interpretStatCommandName(singleCommandName, name, number, modifier);
      m_usingSingleStat = FitStatMethod::createFitStatMethodObject(name, number, modifier);
   }
}

void StatManager::differentiateStats(const IntegerVector& which)
{
   m_alpha.resize(0);
   m_beta.resize(0);

   bool usingChainRule = XSContainer::fit->useChainRule();
   if (usingChainRule) {
     // If ANY of the stats IN USE cannot use the chain rule to evaluate 
     // derivatives then switch to !useChainRule
     // regardless of the setting in the Xspec.init file.
     std::map<size_t, SpectralData*>::const_iterator itSpec = XSContainer::fit->spectraToFit().begin();
     std::map<size_t, SpectralData*>::const_iterator itSpecEnd = XSContainer::fit->spectraToFit().end();
     while ( usingChainRule && itSpec != itSpecEnd ) {
       FitStatMethod* statMeth = getStatMethodForSpectrum(itSpec->first);
       if ( !statMeth->canUseChainRule() ) usingChainRule = false;
       itSpec++;
     }
   }

   usingChainRule ? differentiateUsingChainRule(which) : differentiate(which);
}

void StatManager::setDefaultStatMethod(const string& name, const int& number, const string& modifier)
{
  // ASSUMES that createFitStatMethod either returns a pointer to a new valid object
  // or NULL.
  
  FitStatMethod* tmpStatMethod = 
           FitStatMethod::createFitStatMethodObject(name, number, modifier);
  if (tmpStatMethod) {
    delete m_defaultStatMethod;
    m_defaultStatMethod = tmpStatMethod;
  } else {
    throw YellowAlert("Invalid statistic specifier");
  }
}

FitStatMethod* StatManager::setStatMethodForSpectrum(const size_t specNum, const string& name,
					   const int& number, const string& modifier,
                                           bool storeInManager)
{
  // set the appropriate entry in the member vector
  // Delete any previously existing StatMethod for this spectrum.
  FitStatMethod* fsm = FitStatMethod::createFitStatMethodObject(name, number, modifier);
  if (fsm)
  {
     // find the SpectralData object for this specNum and set the internal information
     SpectralData* spec = XSContainer::datasets->lookup(specNum);
     spec->statCommandName(fsm->commandName());
     
     if (storeInManager)
     {
        std::map<size_t, FitStatMethod*>::iterator itStat = m_fitStatMethods.find(specNum);
        if (itStat != m_fitStatMethods.end())
        {
           delete itStat->second;
           m_fitStatMethods.erase(itStat);
        }
        m_fitStatMethods[specNum] = fsm;
     }

  }
  return fsm;
}


void StatManager::setTestStatMethod(const string& name, 
				    const int& number, const string& modifier)
{
  // ASSUMES that createTestStatMethod either returns a pointer to a new valid object
  // or NULL.
  
  TestStatMethod* tmpStatMethod = 
      TestStatMethod::createTestStatMethodObject(name, number, modifier);
  if (tmpStatMethod) {
     delete m_testStatMethod;
     m_testStatMethod = tmpStatMethod;
  } else {
    throw YellowAlert("Invalid test statistic specifier");
  }
}


std::pair<int,size_t> StatManager::totalDegreesOfFreedom() const
{
  int totDof=0;
  size_t totBins=0;

  std::map<size_t, SpectralData*>::const_iterator itSpec = XSContainer::fit->spectraToFit().begin();
  std::map<size_t, SpectralData*>::const_iterator itSpecEnd = XSContainer::fit->spectraToFit().end();
  while ( itSpec != itSpecEnd ) {
    totBins += getStatMethodForSpectrum(itSpec->first)->nBinsUsed(itSpec->second);
    itSpec++;
  }

  totDof = totBins - XSContainer::fit->variableParameters().size();
  return std::pair<int,size_t>(totDof, totBins);
}

void StatManager::differentiate(const IntegerVector& parNums)
{
   using namespace std;

   const size_t nPars = parNums.size();
   RealArray dPars(.0,nPars);
   m_beta.resize(nPars,0.0);
   for (size_t i=0; i<nPars; ++i)
   {
      ModParam* par = XSContainer::fit->variableParameters(parNums[i]);
      Real origVal = par->value('a');
      Real delta = par->value('d');      
      par->setValue(origVal+delta,'a');      
      performStats();
      Real plusDeltaStat = m_totalStatistic;
      par->setValue(origVal-delta,'a');
      performStats();
      Real minusDeltaStat = m_totalStatistic;
      dPars[i] = delta;
      par->setValue(origVal,'a');
      m_beta[i] = -.25*(plusDeltaStat-minusDeltaStat)/delta;
   }
   if (XSContainer::fit->fitMethod()->secondDerivativeRequired())
   {
      size_t i=0;
      RealArray alphaElems(0.0,nPars*nPars);
      // Calculate:
      // [(S(pi+di,pj+dj)-S(pi-di,pj+dj))-(S(pi+di,pj-dj)-S(pi-di,pj-dj))]/(4di*dj)
      while (i < nPars)
      {
	 ModParam* par_i = XSContainer::fit->variableParameters(parNums[i]);
         const Real orig_i = par_i->value('a');
         const size_t iPos = i*nPars;
         const Real dPars_i = dPars[i];
         for (int iDel=1; iDel>=-1; iDel-=2)
         {
            par_i->setValue(orig_i+iDel*dPars_i,'a');
            size_t j=0;
            while (j <= i)
            {
	       ModParam* par_j = XSContainer::fit->variableParameters(parNums[j]);
               Real orig_j = par_j->value('a');
               for (int jDel=1; jDel>=-1; jDel-=2)
               {
                  par_j->setValue(orig_j+jDel*dPars[j],'a');
                  performStats();
                  const size_t jElem = j+iPos;
                  alphaElems[jElem] += iDel*jDel*m_totalStatistic/(dPars_i*dPars[j]);
                  if (i != j)
                     alphaElems[i + nPars*j] = alphaElems[jElem];
               }
               par_j->setValue(orig_j,'a');
               ++j;
            }
         }
         par_i->setValue(orig_i,'a');
         ++i;
      }
      alphaElems /= 8.0;
      m_alpha.resize(alphaElems.size());
      m_alpha = alphaElems;
   }
   // This will reset all model fluxes as well as statistic.
   performStats();

} // end differentiate

void StatManager::differentiateUsingChainRule(const IntegerVector& parNums)
{        
  // Differentiate the statistic with respect to parameters with
  // indices specified by parNums.

  // NOTE that since each model's parameters are independent of the other
  // models, the result of all this  will be one set of derivative arrays 
  // per model per spectrum.
  using namespace std;

  const size_t nPars = parNums.size();
  // Calculate dsigma/dF arrays for the 2nd term of dStat/dp, which is
  // dStat/dsigma*dsigma/dF*dF/dp.  Since not every StatMethod will need
  // this (due to dStat/dsigma = 0), only do this down at the StatMethod
  // level even though dsigma/dF is itself independent of StatMethod.
  std::map<size_t, SpectralData*>::const_iterator itSpec = XSContainer::fit->spectraToFit().begin();
  std::map<size_t, SpectralData*>::const_iterator itSpecEnd = XSContainer::fit->spectraToFit().end();
  while ( itSpec != itSpecEnd ) {
    getStatMethodForSpectrum(itSpec->first)->getWeightDerivs(itSpec->second);
    itSpec++;
  }

  // Each child process needs to know the CURRENT value of ALL the
  // variable fit parameters, not just the one it's toggling.  But it 
  // only needs to know the delta value for its own.
  vector<double> allVarParVals(nPars);
  for (size_t k=0; k<nPars; ++k) {
    ModParam* par = XSContainer::fit->variableParameters(parNums[k]);
    allVarParVals[k] = par->value('a');
  }
    
  static const int respFloor = Fit::RESPAR_INDEX() << XSContainer::ModelContainer::SHIFT();
  vector<TransferStruct> parallelInput;
  for (size_t k=0; k<nPars; ++k) {
    TransferStruct parInfo;
    ModParam* par = XSContainer::fit->variableParameters(parNums[k]);
    parInfo.iValues.push_back(vector<int>(1, parNums[k]));
    parInfo.dValues.push_back(vector<double>(2));
    parInfo.dValues[0][0] = par->value('a');
    parInfo.dValues[0][1] = par->value('d');
    parInfo.iValues.push_back(parNums);
    parInfo.dValues.push_back(allVarParVals);
    if (parNums[k] > respFloor) {
      // For response params, child processes will need the up-to-date
      // M flux to perform the M*dR/dP calculation.
      const ResponseParam* respPar = dynamic_cast<ResponseParam*>(par);
      if (!respPar) {
	throw RedAlert("Response param missing in differentiateUsingChainRule.\n");
      }
      const Model* modForRespPar = XSContainer::models->lookup(respPar->responseParent());
      const RealArray& currFlux = 
	modForRespPar->modelFlux(respPar->responseParent()->spectrumNumber());
      const size_t nBins = currFlux.size();
      parInfo.dValues.push_back(vector<double>(&currFlux[0], &currFlux[0]+nBins));
    }
    parallelInput.push_back(parInfo);
  }

  // do parallel processing
  ProcessManager::ParallelResults results;
  m_jacobianCalc.run(parallelInput, results);
    
  // check that the output from parallel processing doesn't have major issues
  for (size_t k = 0; k < nPars; ++k) {
    if ( results[k].dValues.size() != XSContainer::fit->spectraToFit().size()) {
      throw RedAlert("Spectrum array size mismatch after parallel derivative calculation.");
    }
  }

  // loop over spectra summing beta and alpha

  m_beta.resize(nPars,0.0);
  bool requireAlpha = XSContainer::fit->fitMethod()->secondDerivativeRequired();
  if ( requireAlpha ) m_alpha.resize(nPars*nPars,0.0);

  itSpec = XSContainer::fit->spectraToFit().begin();
  size_t iSpec = 0;
  while (itSpec != itSpecEnd) {
    const size_t specNum = itSpec->first;
    const size_t nChan = itSpec->second->indirectNotice().size();
    FitStatMethod* stat = getStatMethodForSpectrum(specNum);
    
    // set up map of parameter number and derivative values for this spectrum
    map<size_t,RealArray> dModel;
    for (size_t k = 0; k < nPars; ++k) {
      const size_t parNum = parNums[k];
      const vector<double>& transVec = results[k].dValues[iSpec];
      if (transVec.size() != nChan) {
	throw RedAlert("Spectrum size mismatch after parallel derivative calculation.");
      }
      RealArray& resArray = dModel[parNum];
      resArray.resize(nChan);
      for (size_t i=0; i<nChan; ++i) resArray[i] = transVec[i];
    }

    // add to beta and, if required, alpha for this spectrum

    for (size_t k = 0; k < nPars; ++k) {
      const int i = parNums[k];
      // beta_k is defined as -.5*d(stat)/d(par_k)
      m_beta[k] += -0.5*stat->sumDerivs(dModel[i],specNum,i);
      if ( requireAlpha ) {
	for (size_t m=0; m<=k; m++) {
	  int j = parNums[m];
	  Real alpha_ij = stat->sumSecondDerivs(dModel, i, j);
	  m_alpha[m + nPars*k] += 0.5*alpha_ij;
	  if (k != m) m_alpha[k + nPars*m] += 0.5*alpha_ij;
	}  
      }   
    }

    // end loop over spectra
    itSpec++;
    iSpec++;
  }

  // Now add any required Bayesian prior contribution

  for (size_t k=0; k<nPars; k++) {
    int i = parNums[k];
    m_beta[k] += s_bayes.dlPrior(XSContainer::fit, i);
    if ( requireAlpha ) {
      for (size_t m=0; m<=k; m++) {
	int j = parNums[m];
	m_alpha[m + nPars*k] +=  -s_bayes.d2lPrior(XSContainer::fit, i, j);
	if (k != m) m_alpha[k + nPars*m] += -s_bayes.d2lPrior(XSContainer::fit, j, i);
      }
    }
  }

} // end differentiateUsingChainRule

void StatManager::ParDiff::execute(const bool isParallel, const TransferStruct& input, TransferStruct& output)
{
   using namespace std;

   const int parNum = input.iValues[0][0];
   const Real parVal = input.dValues[0][0];
   const Real delta = input.dValues[0][1];
   
   if (isParallel)
   {
      // Parallel child processes will not have the most recent fit
      // iteration parameter values.  Must set them here.
      const vector<int>& allVarParNums = input.iValues[1];
      const vector<double>& allVarParVals = input.dValues[1];
      for (size_t i=0; i<allVarParNums.size(); ++i)
      {
	 ModParam* modPar = XSContainer::fit->variableParameters(allVarParNums[i]);
         modPar->setValue(allVarParVals[i], 'a');
      }      
   }
   
   ArrayContainer dm;   
   // Model objects contribute to dm ArrayContainers using only "+=" 
   // operations, never "=".  Therefore we must build dm to the
   // proper size here and initialize its arrays to 0.
   std::map<size_t,SpectralData*>::const_iterator itSpec = XSContainer::fit->spectraToFit().begin();
   std::map<size_t,SpectralData*>::const_iterator itSpecEnd = XSContainer::fit->spectraToFit().end();
   while (itSpec != itSpecEnd)
   {
      const size_t nChan = itSpec->second->indirectNotice().size();
      dm.insert(ArrayContainer::value_type(itSpec->first,RealArray(0.,nChan)));
      ++itSpec;
   }
   
   set<string>::const_iterator itName    (XSContainer::fit->activeModels().begin());
   set<string>::const_iterator itNameEnd (XSContainer::fit->activeModels().end());

   // Rather than do a dynamic_cast to tell if this is a
   // ResponseParam, simply rely on the fact that ResponseParams
   // are given an index above a set value.   
   const int respFloor = Fit::RESPAR_INDEX() << XSContainer::ModelContainer::SHIFT();
   ModParam* modp (XSContainer::fit->variableParameters(parNum));
   if (parNum > respFloor)
   {
      ResponseParam* respp = dynamic_cast<ResponseParam*>(modp);
      if (!respp)
      {
         throw RedAlert("Expecting but did not find response parameter in StatMethod.cxx");
      }
      
      // Will need another ArrayContainer for results of a second fold operation.
      ArrayContainer backwardsDm;
      itSpec = XSContainer::fit->spectraToFit().begin();
      while (itSpec != itSpecEnd)
      {
         const size_t nChan = itSpec->second->indirectNotice().size();
         backwardsDm.insert(ArrayContainer::value_type(itSpec->first,RealArray(0.,nChan)));
         ++itSpec;
      }
      
      while (itName != itNameEnd)
      {
         // forward difference  
         // Note that delta is set in unadjusted space so have to go through
         //   some contortions here
         respp->setValue(parVal,'a');
         Real tpar = respp->value('v');
         respp->setValue(tpar + delta,'v');
         // save the model and component fluxes to avoid unnecessary recalculation.
         GroupFluxContainer saveFlux, saveFluxError;
         const string& modelName = *itName;
         if (!isParallel)
         {
	    XSContainer::models->storeFluxes(modelName, saveFlux);
            XSContainer::models->storeFluxVariances(modelName, saveFluxError);
         }
         XSContainer::models->calculate(modelName, true);
         std::pair<XSContainer::ModelMapConstIter,XSContainer::ModelMapConstIter> itRanges 
             = XSContainer::models->modelSet().equal_range(modelName);
         XSContainer::ModelMapConstIter itMod = itRanges.first;
         while (itMod != itRanges.second)
         {
            itMod->second->fold();
            itMod->second->storeDerivative(dm);
            ++itMod;
         }
         
         // backward difference
         respp->setValue(parVal,'a');
         tpar = respp->value('v');
         respp->setValue(tpar - delta,'v');
         XSContainer::models->calculate(modelName);
         itMod = itRanges.first;
         while (itMod != itRanges.second)
         {
            itMod->second->fold();
            itMod->second->storeDerivative(backwardsDm);
            ++itMod;
         }         
         ++itName;
         if (!isParallel)
         {
            XSContainer::models->restoreFluxes (modelName, saveFlux);
            XSContainer::models->restoreFluxVariances(modelName, saveFluxError);
            XSContainer::models->resetComponentFluxes(modelName);
         }
      }  // end model name loop          
      ArrayContainer::iterator itDm = dm.begin();
      ArrayContainer::iterator itBackDm = backwardsDm.begin();
      ArrayContainer::iterator itDmEnd = dm.end();
      while (itDm != itDmEnd)
      {
         itDm->second -= itBackDm->second;
         itDm->second /= (2.0*delta);
         ++itDm;
         ++itBackDm;
      }
      respp->setValue(parVal,'z');
      
   } // end if resp par
   else
   {
      while ( itName != itNameEnd )
      {
         // forward difference  
         const string& modelName = *itName;
         // note that delta is set in unadjusted space so have to go through
         // some contortions here
         modp->setValue(parVal,'a');
         Real tpar = modp->value('v');
         modp->setValue(tpar + delta,'v');
         // save the model and component fluxes to avoid unnecessary recalculation.
         GroupFluxContainer saveFlux, saveFluxError;
         if (!isParallel)
         {
            XSContainer::models->storeFluxes(modelName, saveFlux);
            XSContainer::models->storeFluxVariances(modelName, saveFluxError);
         }
         XSContainer::models->calculate(modelName, true);

         // save results in 'difference' container
         GroupFluxContainer forwardDifference;
         XSContainer::models->storeFluxes(modelName, forwardDifference);

         // backward difference
         modp->setValue(parVal,'a');
         tpar = modp->value('v');
         modp->setValue(tpar - delta,'v');
         XSContainer::models->calculate(modelName);

         // difference contains the forward difference and the models
         // contain the backward difference. Combine these into a derivative,
         // fold, and return the value in dm

         XSContainer::models->makeDerivatives(modelName, forwardDifference, delta, dm);

         // ... and put the model back straight afterwards.
         if (!isParallel)
         {
            XSContainer::models->restoreFluxes (modelName, saveFlux);
            XSContainer::models->restoreFluxVariances(modelName, saveFluxError);
            XSContainer::models->resetComponentFluxes(modelName);
         }

         ++itName;
      } 
      modp->setValue(parVal,'z');
   } // end if regular mod par
   
   
   
   ArrayContainer::iterator itDm = dm.begin();
   ArrayContainer::iterator itDmEnd = dm.end();
   while (itDm != itDmEnd)
   {
      const size_t nChans = itDm->second.size();
      output.dValues.push_back(std::vector<double>(&itDm->second[0],
                                     &itDm->second[0] + nChans));
      ++itDm;
   }
   
}

void StatManager::saveBayes(std::ostream& ostr)
{
   string allSettings;
   if (s_bayes.isOn())
   {
      s_bayes.collectSettings(allSettings, true);
      ostr << "bayes on\n" << allSettings << std::flush;
   }
   else
      ostr << "bayes off\n" << std::flush;
}

void StatManager::printBayes(string& allSettings)
{
   s_bayes.collectSettings(allSettings, false);
}

